@@ -28,6 +28,11 @@ use openpgp::{
     types::HashAlgorithm,
 };
 
+use sequoia_gpg_agent::gnupg;
+
+/// Controls tracing in this module.
+const TRACE: bool = crate::TRACE;
+
 // Setting this to false will cause the creation of a context to fail.
 const ENABLE_GPG: bool = true;
 
@@ -144,9 +149,8 @@ pub fn make_context() -> Result<Ctx> {
 pub fn agent_keypair(policy: &dyn Policy,
                      cert: &Option<Cert>,
                      key: &Key<PublicParts, UnspecifiedRole>)
-                     -> openpgp::Result<sequoia_ipc::gnupg::KeyPair>
+                     -> openpgp::Result<gnupg::KeyPair>
 {
-    use sequoia_ipc::gnupg;
     let agent = gnupg::Context::new()?;
     let mut s = gnupg::KeyPair::new(&agent, key)?;
     if let Some(cert) = cert.as_ref() {
@@ -454,9 +458,9 @@ pub fn import(ctx: &Ctx, what: &[u8]) -> Result<()> {
 }
 
 pub fn export(ctx: &Ctx, search: Option<&str>, if_changed: Option<&CacheTag>)
-    -> Result<(String, CacheTag)>
+    -> Result<(Vec<u8>, CacheTag)>
 {
-    let doit = || -> Result<(String, CacheTag)> {
+    let doit = || -> Result<(Vec<u8>, CacheTag)> {
         let tag = CacheTag::new(ctx)?;
         if let Some(if_changed) = if_changed {
             if tag.metadata_eq(if_changed) {
@@ -473,7 +477,6 @@ pub fn export(ctx: &Ctx, search: Option<&str>, if_changed: Option<&CacheTag>)
             .stderr(Stdio::null())
             .arg("--homedir")
             .arg(ctx.directory("homedir")?)
-            .arg("--armor")
             .arg("--export-options")
             .arg("export-local-sigs")
             .arg("--export");
@@ -491,7 +494,7 @@ pub fn export(ctx: &Ctx, search: Option<&str>, if_changed: Option<&CacheTag>)
             }
         }
 
-        Ok((String::from_utf8_lossy(&gpg.stdout).into_owned(), tag))
+        Ok((gpg.stdout, tag))
     };
 
     Ok(doit()?)
@@ -726,7 +729,7 @@ pub fn list_validity(ctx: &Ctx, filter: Option<String>,
                      if_changed: Option<&CacheTag>)
     -> Result<(Vec<(Fingerprint, Vec<(String, Validity)>)>, CacheTag)>
 {
-    rnp_function!(gpg::list_validity, super::TRACE);
+    rnp_function!(gpg::list_validity, TRACE);
 
     let (res, tag) = list_keys_raw(&ctx, false, filter, if_changed)?;
 
@@ -807,7 +810,7 @@ pub fn list_secret_keys(ctx: &Ctx, filter: Option<String>,
                         if_changed: Option<&CacheTag>)
     -> Result<(Vec<(Fingerprint, Fingerprint)>, CacheTag)>
 {
-    rnp_function!(gpg::list_secret_keys, super::TRACE);
+    rnp_function!(gpg::list_secret_keys, TRACE);
 
     let (res, tag) = list_keys_raw(&ctx, true, filter, if_changed)?;
     let secret_keys = list_secret_keys_parse(res)?;
@@ -818,7 +821,7 @@ pub fn list_secret_keys(ctx: &Ctx, filter: Option<String>,
 pub fn list_secret_keys_parse(res: Vec<csv::StringRecord>)
     -> Result<Vec<(Fingerprint, Fingerprint)>>
 {
-    rnp_function!(gpg::list_secret_keys_parse, super::TRACE);
+    rnp_function!(gpg::list_secret_keys_parse, TRACE);
 
     let mut secret_keys: Vec<(Fingerprint, Fingerprint)> = Vec::new();
 
