@@ -28,6 +28,8 @@ use std::{
     },
 };
 
+use anyhow::Result;
+
 use sequoia_openpgp as openpgp;
 use openpgp::{
     Fingerprint,
@@ -42,9 +44,12 @@ use crate::{
     gpg,
     Keygrip,
     parcimonie::Parcimonie,
-    wot::WoT,
+    wot::{WoT, WoTWorkerHandle},
 };
 
+/// Controls tracing in this module.
+const TRACE: bool = crate::TRACE;
+
 #[derive(Debug)]
 struct MapEntry<E: std::fmt::Debug> {
     cert: Arc<RwLock<Cert>>,
@@ -127,14 +132,14 @@ pub struct KeystoreData {
     // store a reference.  Right now that function is only used
     // outside of the Keystore to lock and unlock secret key material.
     // But, since this is just a safety measure, this is good enough.
-    last_update: Arc<AtomicUsize>,
+    last_update: AtomicUsize,
 
     // The location of the keystore.  If set, the keystore is
     // periodically flushed to disk in the background.
     directory: Option<PathBuf>,
     flush_thread: Option<thread::JoinHandle<()>>,
 
-    gpg_thread: Option<thread::JoinHandle<()>>,
+    background_thread: Option<thread::JoinHandle<()>>,
 }
 
 impl Default for KeystoreData {
@@ -153,12 +158,12 @@ impl Default for KeystoreData {
             keys_on_agent_last_refresh: std::time::UNIX_EPOCH,
             keys_on_agent_tag: None,
 
-            last_update: Arc::new(AtomicUsize::new(0)),
+            last_update: AtomicUsize::new(0),
 
             directory: None,
             flush_thread: None,
 
-            gpg_thread: None,
+            background_thread: None,
         }
     }
 }
@@ -167,22 +172,28 @@ pub struct Keystore {
     data: Arc<RwLock<KeystoreData>>,
     parcimonie: Option<Parcimonie>,
 
-    // The wot functionality gets enabled for the 'main' context, not for
-    // ephemeral RnpContexts. Wot calculations are only done for the 'main'
-    // context.
-    //
-    // We assume rnp_load_keys() is only used by the 'main' context, so wot
-    // gets initialized there.
-    wot : Option<WoT>,
-
-    // A sender to ping the thread and tell it to check for updates
-    // now.
-    gpg_sender: Option<Arc<Mutex<mpsc::Sender<Option<(Vec<u8>, bool)>>>>>,
+    /// Write-side of a channel to send commands to the background thread.
+    background_sender:
+    Option<Arc<Mutex<mpsc::Sender<BackgroundCommand>>>>,
 
     // Whether the keystore is in the process of loading a keyring.
     // See Keystore::load_keyring_in_background and
     // Keystore::block_on_load.
-    in_init: Arc<(Mutex<usize>, Condvar)>,
+    background_busy: Arc<(Mutex<usize>, Condvar)>,
+
+    /// Controls the WoT worker thread.
+    wot_handle: Option<WoTWorkerHandle>,
+}
+
+enum BackgroundCommand {
+    /// Update the GnuPG keyring.
+    LoadGnuPGKeyring,
+
+    /// Import a keyring in the background.
+    LoadKeyring {
+        keyring: Vec<u8>,
+        import_secret_keys: bool,
+    },
 }
 
 impl Default for Keystore {
@@ -191,10 +202,9 @@ impl Default for Keystore {
             data: Arc::new(RwLock::new(Default::default())),
             parcimonie: None,
 
-            wot: None,
-
-            gpg_sender: None,
-            in_init: Arc::new((Mutex::new(0), Condvar::new())),
+            background_sender: None,
+            background_busy: Arc::new((Mutex::new(0), Condvar::new())),
+            wot_handle: None,
         }
     }
 }
@@ -213,32 +223,12 @@ impl Keystore {
     //
     // If there is already parcimonie instance running, this does
     // nothing.
-    pub fn start_parcimonie(&mut self, policy: StandardPolicy<'static>) {
+    pub fn start_parcimonie(&mut self, policy: StandardPolicy<'static>)
+                            -> Result<()>
+    {
         if self.parcimonie.is_none() {
-            self.parcimonie = Some(Parcimonie::new(policy, self));
+            self.parcimonie = Some(Parcimonie::new(policy, self)?);
         }
-    }
-
-    pub fn enable_wot(&mut self, policy: &StandardPolicy<'static>)
-        -> openpgp::Result<()>
-    {
-        let mut wot = WoT::new()?;
-        wot.update(self.create_ref(), policy)?;
-
-        self.wot = Some(wot);
-
-        Ok(())
-    }
-
-    pub fn update_wot(&mut self, policy: &StandardPolicy<'static>)
-        -> openpgp::Result<()>
-    {
-        let ksd = self.create_ref();
-
-        if let Some(ref mut wot) = self.wot {
-            wot.update(ksd, policy)?;
-        }
-
         Ok(())
     }
 
@@ -270,7 +260,7 @@ impl Keystore {
     // The location of the keystore (where the pubring.gpg and
     // secring.gpg files are).  If set, the keystore is periodically
     // flushed to disk in the background.
-    pub fn set_directory(&mut self, directory: PathBuf) {
+    pub fn set_directory(&mut self, directory: PathBuf) -> Result<()> {
         let mut ks = self.data.write().unwrap();
 
         ks.directory = Some(directory);
@@ -280,52 +270,62 @@ impl Keystore {
             // disk.
             let ks_ref = self.create_ref();
             ks.flush_thread
-                = Some(thread::spawn(move || {
+                = Some(thread::Builder::new()
+                       .name("sq flusher".to_string())
+                       .spawn(move || {
                     KeystoreData::flush_thread(ks_ref);
-                }));
+                })?);
         }
+        Ok(())
     }
 
     // Start the thread that checks for updates to the gpg key store.
     //
     // Returns whether or not a thread was started.  That is, if the
     // thread was already running, this returns false.
-    fn gpg_thread_start(&mut self,
+    fn background_threads_start(&mut self,
                         policy: Arc<RwLock<StandardPolicy<'static>>>)
         -> openpgp::Result<bool>
     {
-        if self.gpg_sender.is_some() {
+        if self.background_sender.is_some() {
             // Already started.
             return Ok(false);
         }
 
         let mut ks = self.data.write().unwrap();
-        if self.gpg_sender.is_some() {
+        if self.background_sender.is_some() {
             // Someone else already started the thread while we were
             // blocked on self.data.
-            assert!(ks.gpg_thread.is_some());
+            assert!(ks.background_thread.is_some());
             return Ok(false);
         }
-        assert!(ks.gpg_thread.is_none());
+        assert!(ks.background_thread.is_none());
+        assert!(self.background_sender.is_none());
+        assert!(self.wot_handle.is_none());
+
+        let wot_handle =
+            WoT::background_thread_start(self.create_ref(), policy.clone())?;
 
         let ks_ref = self.create_ref();
         let (sender, receiver) = mpsc::channel();
 
-        let in_init = Arc::clone(&self.in_init);
+        let background_busy = Arc::clone(&self.background_busy);
 
         // Cause block_on_load to wait until the gpg keyring is
         // loaded.
-        let mut in_init_ = self.in_init.0.lock().unwrap();
-        assert_eq!(*in_init_, 0);
-        *in_init_ += 1;
-        drop(in_init_);
+        let mut background_busy_ = self.background_busy.0.lock().unwrap();
+        assert_eq!(*background_busy_, 0);
+        *background_busy_ += 1;
+        drop(background_busy_);
 
-        ks.gpg_thread
-            = Some(thread::spawn(move || {
-                KeystoreData::gpg_thread(ks_ref, receiver, in_init, policy);
-            }));
+        let wh = wot_handle.clone();
+        ks.background_thread
+            = Some(thread::Builder::new().name("sq loader".to_string())
+                   .spawn(move || KeystoreData::background_thread(
+                       ks_ref, wh,receiver, background_busy))?);
 
-        self.gpg_sender = Some(Arc::new(Mutex::new(sender)));
+        self.background_sender = Some(Arc::new(Mutex::new(sender)));
+        self.wot_handle = Some(wot_handle);
 
         Ok(true)
     }
@@ -338,19 +338,19 @@ impl Keystore {
     ///
     /// This also starts a thread to monitor the gpg store.  This
     /// imports updates from the gpg keyring (XXX: but not removals!).
-    pub fn load_gpg_keyring(&mut self,
+    pub fn load_gpg_keyring_in_background(&mut self,
                             policy: Arc<RwLock<StandardPolicy<'static>>>)
         -> openpgp::Result<()>
     {
-        rnp_function!(Keystore::load_gpg_keyring, super::TRACE);
+        rnp_function!(Keystore::load_gpg_keyring, TRACE);
 
-        if !self.gpg_thread_start(policy)? {
+        if !self.background_threads_start(policy)? {
             // If we just started the thread, then it will read the
             // gpg keyring on its own.  Otherwise, signal that it
             // should reread it now.
-            self.gpg_sender.as_ref().expect("started thread")
+            self.background_sender.as_ref().expect("started thread")
                 .lock().unwrap()
-                .send(None)?;
+                .send(BackgroundCommand::LoadGnuPGKeyring)?;
         }
         Ok(())
     }
@@ -364,23 +364,38 @@ impl Keystore {
                                       policy: Arc<RwLock<StandardPolicy<'static>>>)
         -> openpgp::Result<()>
     {
-        rnp_function!(Keystore::load_keyring_in_background, super::TRACE);
+        rnp_function!(Keystore::load_keyring_in_background, TRACE);
 
-        if let Err(err) = self.gpg_thread_start(policy) {
+        if let Err(err) = self.background_threads_start(policy) {
             warn!("Failed to start gpg thread: {}", err);
             KeystoreData::load_keyring(
                 &mut self.data, &keyring, import_secret_keys, false)?;
         } else {
-            *self.in_init.0.lock().unwrap() += 1;
+            *self.background_busy.0.lock().unwrap() += 1;
 
-            self.gpg_sender.as_ref().expect("started thread")
+            self.background_sender.as_ref().expect("started thread")
                 .lock().unwrap()
-                .send(Some((keyring, import_secret_keys)))?;
+                .send(BackgroundCommand::LoadKeyring {
+                    keyring,
+                    import_secret_keys,
+                })?;
         }
 
         Ok(())
     }
 
+    /// Updates the Web-of-Trust information in the background.
+    pub fn update_wot_in_background(&mut self,
+                                    policy: Arc<RwLock<StandardPolicy<'static>>>)
+                                    -> openpgp::Result<()> {
+        rnp_function!(Keystore::update_wot_in_background, super::TRACE);
+
+        self.background_threads_start(policy)?;
+        self.wot_handle.as_ref().expect("started thread").notify();
+
+        Ok(())
+    }
+
     /// Blocks the current thread until any keyrings that are being
     /// loaded in the background have been loaded.
     ///
@@ -391,10 +406,10 @@ impl Keystore {
     pub fn block_on_load(&mut self) -> openpgp::Result<bool> {
         // We do *not* start the gpg thread if it is not already
         // running.  This must only be done by rnp_load_keys.
-        rnp_function!(Keystore::block_on_load, super::TRACE);
+        rnp_function!(Keystore::block_on_load, TRACE);
 
-        let in_init = self.in_init.0.lock().unwrap();
-        if *in_init == 0 {
+        let background_busy = self.background_busy.0.lock().unwrap();
+        if *background_busy == 0 {
             return Ok(false);
         }
 
@@ -441,10 +456,10 @@ impl Keystore {
             }
         };
 
-        assert!(self.gpg_sender.is_some());
+        assert!(self.background_sender.is_some());
 
-        drop(self.in_init.1
-             .wait_while(in_init, |in_init| *in_init > 0)
+        drop(self.background_busy.1
+             .wait_while(background_busy, |background_busy| *background_busy > 0)
              .unwrap());
 
         if let Some(mut child) = child {
@@ -533,9 +548,20 @@ impl fmt::Debug for KeystoreData {
 }
 
 impl KeystoreData {
+    /// Returns the time (in ticks) where the keystore could have
+    /// possibly been modified.
+    pub fn last_update(&self) -> usize {
+        self.last_update.load(atomic::Ordering::Relaxed)
+    }
+
+    /// Marks the keystore as updated.
+    pub fn mark_updated(&self) {
+        self.last_update.fetch_add(1, atomic::Ordering::Relaxed);
+    }
+
     // Periodically flush the key store to disk.
     fn flush_thread(ks: Arc<RwLock<Self>>) {
-        rnp_function!(KeystoreData::flush_thread, super::TRACE);
+        rnp_function!(KeystoreData::flush_thread, TRACE);
         let mut last_flush = 0;
 
         loop {
@@ -545,7 +571,7 @@ impl KeystoreData {
 
             let ks = ks.read().unwrap();
 
-            let last_update = ks.last_update.load(atomic::Ordering::Relaxed);
+            let last_update = ks.last_update();
             if last_flush == last_update {
                 t!("Not flushing keystore to disk: no changes");
                 continue;
@@ -655,7 +681,7 @@ impl KeystoreData {
                     keyring: &[u8], import_secret_keys: bool, external: bool)
         -> openpgp::Result<()>
     {
-        rnp_function!(KeystoreData::load_keyring, super::TRACE);
+        rnp_function!(KeystoreData::load_keyring, TRACE);
 
         // Parse the keyring without the lock.
         use sequoia_openpgp_mt::keyring;
@@ -698,40 +724,46 @@ impl KeystoreData {
         Ok(())
     }
 
-    // Monitor gpg for updates.
-    //
-    // This checks for updates to the keyring.
-    fn gpg_thread(ks: Arc<RwLock<Self>>,
-                  receiver: mpsc::Receiver<Option<(Vec<u8>, bool)>>,
-                  in_init: Arc<(Mutex<usize>, Condvar)>,
-                  policy: Arc<RwLock<StandardPolicy<'static>>>) {
-        rnp_function!(KeystoreData::gpg_thread, super::TRACE);
+    /// The background worker thread.
+    ///
+    /// Monitor certificate stores (like the GnuPG keyring) for
+    /// updates, reads certificates in the background, and updates the
+    /// Web-of-Trust calculations.
+    fn background_thread(ks: Arc<RwLock<Self>>,
+                         wot_handle: WoTWorkerHandle,
+                         receiver: mpsc::Receiver<BackgroundCommand>,
+                         background_busy: Arc<(Mutex<usize>, Condvar)>) {
+        rnp_function!(KeystoreData::background_thread, TRACE);
 
-        let mut keyring_tag: Option<gpg::CacheTag> = None;
-        let mut gpg_ctx: Option<gpg::Ctx> = None;
+        struct GpgState {
+            last_check: std::time::SystemTime,
+            first_time: bool,
+            check_now: bool,
+            keyring_tag: Option<gpg::CacheTag>,
+            ctx: Option<gpg::Ctx>,
+        }
 
-        let mut wot = match WoT::new() {
-            Ok(wot) => Some(wot),
-            Err(_e) => None, // FIXME
+        let mut gpg = GpgState {
+            last_check: std::time::UNIX_EPOCH,
+            first_time: true,
+            check_now: true,
+            keyring_tag: None,
+            ctx: None,
         };
-
-        let mut last_check = std::time::UNIX_EPOCH;
-        let mut first_time = true;
-        let mut check_now = true;
         loop {
             // About once an hour (but not quite to avoid possible
             // resonance).  The first time through, we don't import
             // the keyring.  We did that from the main thread, because
             // reasons.
-            let timeout = if check_now {
+            let timeout = if gpg.check_now {
                 Duration::new(0, 0)
             } else {
                 let threshold = Duration::new(61 * 60, 0);
-                let elapsed = SystemTime::now().duration_since(last_check)
+                let elapsed = SystemTime::now().duration_since(gpg.last_check)
                     .unwrap_or_else(|_| {
                         // Clock skew?  Who knows.  Just wait a bit
                         // and try again.
-                        last_check = SystemTime::now();
+                        gpg.last_check = SystemTime::now();
                         Duration::new(0, 0)
                     });
                 if elapsed > threshold {
@@ -747,27 +779,31 @@ impl KeystoreData {
 
             let r = receiver.recv_timeout(timeout);
             match r {
-                // Import a keyring in the background.
-                Ok(Some((keyring, import_secret_keys))) => {
+                Ok(BackgroundCommand::LoadKeyring {
+                    keyring,
+                    import_secret_keys,
+                }) => {
                     if let Err(err) = KeystoreData::load_keyring(
                         &ks, &keyring, import_secret_keys, false)
                     {
                         warn!("Loading keyring: {}", err);
                     }
 
-                    let mut in_init_ = in_init.0.lock().unwrap();
-                    assert!(*in_init_ > 0);
-                    *in_init_ -= 1;
-                    if *in_init_ == 0 {
-                        in_init.1.notify_all();
+                    let mut background_busy_ = background_busy.0.lock().unwrap();
+                    assert!(*background_busy_ > 0);
+                    *background_busy_ -= 1;
+                    if *background_busy_ == 0 {
+                        background_busy.1.notify_all();
                     }
-                    drop(in_init_);
+                    drop(background_busy_);
 
+                    // Trigger a WoT update.
+                    wot_handle.notify();
                     continue;
                 }
-                // Update the keyring now.
-                Ok(None) => {
-                    check_now = true;
+
+                Ok(BackgroundCommand::LoadGnuPGKeyring) => {
+                    gpg.check_now = true;
                     // We loop to drain the message queue.
                     continue;
                 },
@@ -780,10 +816,10 @@ impl KeystoreData {
                 }
             }
 
-            check_now = false;
+            gpg.check_now = false;
 
             let now = SystemTime::now();
-            let elapsed = now.duration_since(last_check);
+            let elapsed = now.duration_since(gpg.last_check);
             match elapsed {
                 Err(err) => {
                     t!("Error computing time since last update: {}", err);
@@ -791,31 +827,31 @@ impl KeystoreData {
                 },
                 Ok(elapsed) => {
                     if elapsed < Duration::new(90, 0) {
-                        t!("Throttling gpg_thread.");
+                        t!("Throttling GnuPG certstore updates.");
                         continue;
                     }
                 }
             }
 
-            if gpg_ctx.is_none() {
+            if gpg.ctx.is_none() {
                 // See if the user install gpg in the meantime.
-                gpg_ctx = gpg::Ctx::new().ok();
+                gpg.ctx = gpg::Ctx::new().ok();
             }
-            if let Some(gpg_ctx) = gpg_ctx.as_ref() {
+            if let Some(gpg_ctx) = gpg.ctx.as_ref() {
                 t!("Checking for gpg updates (last check was: {:?} ago)...",
                    elapsed);
 
                 // Import the gpg keyring.
                 match gpg::export(
-                    gpg_ctx, None, keyring_tag.as_ref())
+                    gpg_ctx, None, gpg.keyring_tag.as_ref())
                 {
                     Ok((keyring, keyring_tag_tmp)) => {
                         t!("Got {} bytes", keyring.len());
-                        keyring_tag = Some(keyring_tag_tmp);
+                        gpg.keyring_tag = Some(keyring_tag_tmp);
 
 
                         if let Err(err) = KeystoreData::load_keyring(
-                            &ks, keyring.as_bytes(), false, true)
+                            &ks, &keyring, false, true)
                         {
                             warn!("Loading keyring: {}", err);
                         }
@@ -826,31 +862,28 @@ impl KeystoreData {
                 }
             }
 
-            if first_time {
-                first_time = false;
+            if gpg.first_time {
+                gpg.first_time = false;
 
-                let mut in_init_ = in_init.0.lock().unwrap();
-                assert!(*in_init_ > 0);
-                *in_init_ -= 1;
-                if *in_init_ == 0 {
-                    in_init.1.notify_all();
+                let mut background_busy_ = background_busy.0.lock().unwrap();
+                assert!(*background_busy_ > 0);
+                *background_busy_ -= 1;
+                if *background_busy_ == 0 {
+                    background_busy.1.notify_all();
                 }
-                drop(in_init_);
+                drop(background_busy_);
             }
 
-            // perform WoT update
-            if let Some(ref mut wot) = wot {
-                let p = policy.read().unwrap();
-                wot.update(ks.clone(), &*p).expect("foo");
-            }
+            // Trigger a WoT update.
+            wot_handle.notify();
 
-            last_check = SystemTime::now();
+            gpg.last_check = SystemTime::now();
         }
     }
 
     /// Returns a reference to a gpg::Ctx.
     fn gpg_ctx(&mut self) -> Option<&gpg::Ctx> {
-        rnp_function!(KeystoreData::gpg_ctx, super::TRACE);
+        rnp_function!(KeystoreData::gpg_ctx, TRACE);
 
         if self.gpg_ctx.is_none() {
             match gpg::Ctx::new() {
@@ -866,7 +899,7 @@ impl KeystoreData {
     ///
     /// This always tries to refresh the cache first.
     pub fn key_on_agent_hard(&mut self, key: &Fingerprint) -> bool {
-        rnp_function!(KeystoreData::key_on_agent_hard, super::TRACE);
+        rnp_function!(KeystoreData::key_on_agent_hard, TRACE);
 
         self.gpg_ctx();
         if let Some(ctx) = self.gpg_ctx.as_ref() {
@@ -910,7 +943,7 @@ impl KeystoreData {
     /// The calls are split so that this function doesn't need a write
     /// lock.
     pub fn key_on_agent(&self, key: &Fingerprint) -> (bool, bool) {
-        rnp_function!(KeystoreData::key_on_agent, super::TRACE);
+        rnp_function!(KeystoreData::key_on_agent, TRACE);
 
         let now = SystemTime::now();
         let elapsed = now.duration_since(self.keys_on_agent_last_refresh);
@@ -958,11 +991,26 @@ impl KeystoreData {
     /// Returns a clone of the specified certificate's RefCell.
     ///
     /// fp is the certificate's fingerprint.
-    pub fn cert_cell(&self, fp: &Fingerprint) -> Option<Arc<RwLock<Cert>>> {
-        self.last_update.fetch_add(1, atomic::Ordering::Relaxed);
+    pub fn cert_cell_by_primary_fp(&self, fp: &Fingerprint)
+                                   -> Option<Arc<RwLock<Cert>>>
+    {
         self.by_primary_fp.get(fp).map(|e| e.cert_cell())
     }
 
+    /// Returns a clone of the specified certificates' storage cells.
+    ///
+    /// `fp` is a certificate or subkey fingerprint.
+    pub fn cert_cell_by_fp<'a>(&'a self, fp: &Fingerprint)
+                               -> impl Iterator<Item = Arc<RwLock<Cert>>> + 'a
+    {
+        std::iter::once(fp.clone())
+            .chain(self.by_subkey_fp.get(fp)
+                   .unwrap_or(VEC_MAP_ENTRY_EMPTY_REF)
+                   .iter()
+                   .map(|e| e.fingerprint().clone()))
+            .filter_map(|fp| self.by_primary_fp.get(&fp).map(|e| e.cert_cell()))
+    }
+
     /// Returns a reference to the cert with the given fingerprint, if
     /// any.
     pub fn by_primary_fp(&self, fp: &Fingerprint)
@@ -1065,7 +1113,7 @@ impl KeystoreData {
     /// This function prefers secret key material in the cert that is
     /// being merged.
     fn insert_(&mut self, cert: Cert, is_external: bool) {
-        rnp_function!(Keystore::insert_, super::TRACE);
+        rnp_function!(Keystore::insert_, TRACE);
 
         let fp = cert.fingerprint();
         let id = KeyID::from(&fp);
@@ -1138,7 +1186,7 @@ impl KeystoreData {
             }
         };
 
-        self.last_update.fetch_add(1, atomic::Ordering::Relaxed);
+        self.mark_updated();
 
         // If this is an update, then the primary key id and primary
         // key grip mapping exist and are correct.
@@ -1309,7 +1357,7 @@ impl KeystoreData {
         }
 
         if removed_one {
-            self.last_update.fetch_add(1, atomic::Ordering::Relaxed);
+            self.mark_updated();
         }
 
         removed_one
@@ -1359,7 +1407,7 @@ impl KeystoreData {
         }
 
         if removed_one {
-            self.last_update.fetch_add(1, atomic::Ordering::Relaxed);
+            self.mark_updated();
         }
 
         removed_one
@@ -1692,7 +1740,8 @@ mod tests {
 
         // Merge the revocation certificate.
         {
-            let cell = ks.cert_cell(&alice.fingerprint()).unwrap();
+            let cell =
+                ks.cert_cell_by_primary_fp(&alice.fingerprint()).unwrap();
             let mut cert = cell.write().unwrap();
             *cert = cert.clone().insert_packets(Some(revocation))?;
         }
