@@ -4,9 +4,6 @@ use std::{
     collections::{
         HashMap,
     },
-    ffi::{
-        CStr,
-    },
     sync::atomic,
 };
 use std::sync::{Arc, RwLock, RwLockReadGuard};
@@ -186,12 +183,10 @@ fn rnp_ffi_create(ctx: *mut *mut RnpContext,
 {
     rnp_function!(rnp_ffi_create, crate::TRACE);
     assert_ptr!(ctx);
-    assert_ptr!(pub_fmt);
-    assert_ptr!(sec_fmt);
-    if CStr::from_ptr(pub_fmt).to_bytes() != b"GPG"
-        || CStr::from_ptr(sec_fmt).to_bytes() != b"GPG"
-    {
-        return RNP_ERROR_BAD_FORMAT;
+    let pub_fmt = assert_str!(pub_fmt);
+    let sec_fmt = assert_str!(sec_fmt);
+    if pub_fmt != "GPG" || sec_fmt != "GPG" {
+        rnp_return_status!(RNP_ERROR_BAD_FORMAT);
     }
 
     // Try to make sure that a pubring.gpg exists.  Thunderbird will
@@ -262,22 +257,24 @@ fn rnp_ffi_create(ctx: *mut *mut RnpContext,
         policy: Arc::new(RwLock::new(policy)),
         ..Default::default()
     }));
-    RNP_SUCCESS
+    rnp_success!()
 }
 
 #[no_mangle] pub unsafe extern "C"
 fn rnp_ffi_destroy(ctx: *mut RnpContext) -> RnpResult {
+    rnp_function!(rnp_ffi_destroy, crate::TRACE);
+    arg!(ctx);
     if ! ctx.is_null() {
         drop(Box::from_raw(ctx));
     }
-    RNP_SUCCESS
+    rnp_success!()
 }
 
 #[no_mangle] pub unsafe extern "C"
 fn rnp_ffi_set_log_fd(ctx: *mut RnpContext, _fd: c_int) -> RnpResult {
     rnp_function!(rnp_ffi_set_log_fd, crate::TRACE);
-    assert_ptr!(ctx);
-    RNP_SUCCESS
+    let _ = assert_ptr_mut!(ctx);
+    rnp_success!()
 }
 
 #[no_mangle] pub unsafe extern "C"
@@ -286,9 +283,11 @@ fn rnp_ffi_set_pass_provider(ctx: *mut RnpContext,
                              cookie: *mut c_void)
                              -> RnpResult {
     rnp_function!(rnp_ffi_set_pass_provider, crate::TRACE);
-    assert_ptr!(ctx);
-    (*ctx).password_cb = Some((cb, cookie));
-    RNP_SUCCESS
+    let ctx = assert_ptr_mut!(ctx);
+    arg!(cb);
+    arg!(cookie);
+    ctx.password_cb = Some((cb, cookie));
+    rnp_success!()
 }
 
 impl RnpContext {
@@ -626,9 +625,10 @@ fn rnp_load_keys(ctx: *mut RnpContext,
                  -> RnpResult {
     rnp_function!(rnp_load_keys, TRACE);
 
-    assert_ptr!(ctx);
-    assert_ptr!(format);
-    assert_ptr!(input);
+    let ctx = assert_ptr_mut!(ctx);
+    let format = assert_str!(format);
+    let input = assert_ptr_mut!(input);
+    arg!(flags);
 
     static BANNER_SHOWN: atomic::AtomicBool = atomic::AtomicBool::new(false);
     if ! BANNER_SHOWN.load(atomic::Ordering::Relaxed) {
@@ -645,12 +645,10 @@ fn rnp_load_keys(ctx: *mut RnpContext,
         BANNER_SHOWN.store(true, atomic::Ordering::Relaxed);
     }
 
-
-    if CStr::from_ptr(format).to_bytes() != b"GPG" {
-        return RNP_ERROR_BAD_FORMAT;
+    if format != "GPG" {
+        rnp_return_status!(RNP_ERROR_BAD_FORMAT);
     }
 
-    let input = &mut *input;
     let input_size = input.size();
 
     match flags {
@@ -665,7 +663,8 @@ fn rnp_load_keys(ctx: *mut RnpContext,
                         if input_size == file_size {
                             t!("Looks like a match.  Periodically flushing \
                                 the keystore to disk.");
-                            (*ctx).certs.set_directory(profile);
+                            rnp_try_or!((*ctx).certs.set_directory(profile),
+                                        RNP_ERROR_GENERIC);
                         } else {
                             t!("pubring.gpg does not match input.  \
                                 Conservatively disabling flushing the \
@@ -676,22 +675,20 @@ fn rnp_load_keys(ctx: *mut RnpContext,
             }
 
             // Also load GPG's public key database.
-            if let Err(err) = (*ctx).certs.load_gpg_keyring((*ctx).policy.clone()) {
+            if let Err(err) = (*ctx).certs.load_gpg_keyring_in_background(
+                (*ctx).policy.clone())
+            {
                 warn!("Import gpg's keyring: {}", err);
             }
 
-            // Enable the WoT functionality.
-            if let Err(err) = (*ctx).certs.enable_wot(&(*ctx).policy()) {
-                warn!("Enabling the WoT functionality: {}", err);
-            }
-
-            (*ctx).certs.start_parcimonie((*ctx).policy().clone());
+            let policy = ctx.policy().clone();
+            rnp_try_or!(ctx.certs.start_parcimonie(policy), RNP_ERROR_GENERIC);
         }
         RNP_LOAD_SAVE_SECRET_KEYS => (),
         f => {
             warn!("sequoia-octopus: unexpected flags to rnp_load_keys: {:x}",
                   f);
-            return RNP_ERROR_BAD_PARAMETERS;
+            rnp_return_status!(RNP_ERROR_BAD_PARAMETERS);
         },
     }
 
@@ -700,7 +697,7 @@ fn rnp_load_keys(ctx: *mut RnpContext,
     if let Err(err) = input.read_to_end(&mut data)
     {
         warn!("sequoia-octopus: Error reading input: {}", err);
-        return RNP_ERROR_GENERIC;
+        rnp_return_status!(RNP_ERROR_GENERIC);
     }
 
     let policy = (*ctx).policy.clone();
@@ -708,10 +705,10 @@ fn rnp_load_keys(ctx: *mut RnpContext,
         data, flags == RNP_LOAD_SAVE_SECRET_KEYS, policy)
     {
         warn!("sequoia-octopus: Error reading certs: {}", err);
-        return RNP_ERROR_GENERIC;
+        rnp_return_status!(RNP_ERROR_GENERIC);
     }
 
-    RNP_SUCCESS
+    rnp_success!()
 }
 
 #[no_mangle] pub unsafe extern "C"
@@ -721,14 +718,13 @@ fn rnp_save_keys(ctx: *mut RnpContext,
                  flags: RnpLoadSaveFlags)
                  -> RnpResult {
     rnp_function!(rnp_save_keys, TRACE);
-    assert_ptr!(ctx);
-    assert_ptr!(format);
-    assert_ptr!(output);
-    if CStr::from_ptr(format).to_bytes() != b"GPG" {
-        return RNP_ERROR_BAD_FORMAT;
+    let ctx = assert_ptr_mut!(ctx);
+    let format = assert_str!(format);
+    let output = assert_ptr_mut!(output);
+    if format != "GPG" {
+        rnp_return_status!(RNP_ERROR_BAD_FORMAT);
     }
 
-    let output = &mut *output;
     let mut r = Ok(());
     let mut count = 0;
     match flags {
@@ -756,7 +752,7 @@ fn rnp_save_keys(ctx: *mut RnpContext,
         }
         f => {
             warn!("unexpected flags to rnp_save_keys: {:x}", f);
-            return RNP_ERROR_BAD_PARAMETERS;
+            rnp_return_status!(RNP_ERROR_BAD_PARAMETERS);
         },
     };
 
@@ -775,12 +771,12 @@ fn rnp_save_keys(ctx: *mut RnpContext,
         }
     }
 
-    if let Err(err) = r {
+    rnp_return_status!(if let Err(err) = r {
         warn!("failed saving keys: {}", err);
         RNP_ERROR_GENERIC
     } else {
         RNP_SUCCESS
-    }
+    })
 }
 
 #[no_mangle] pub unsafe extern "C"
@@ -789,6 +785,7 @@ fn rnp_get_public_key_count(ctx: *mut RnpContext,
                             -> RnpResult {
     rnp_function!(rnp_get_public_key_count, crate::TRACE);
     let ctx = assert_ptr_mut!(ctx);
+    let count = assert_ptr_mut!(count);
 
     // We load the keyrings in the background.  But, once TB tries to
     // get the number of certs, we have to wait on that process.
@@ -811,7 +808,7 @@ fn rnp_get_public_key_count(ctx: *mut RnpContext,
         }
     }).count();
     t!("-> {}", *count);
-    RNP_SUCCESS
+    rnp_success!()
 }
 
 #[no_mangle] pub unsafe extern "C"
@@ -820,6 +817,7 @@ fn rnp_get_secret_key_count(ctx: *mut RnpContext,
                             -> RnpResult {
     rnp_function!(rnp_get_secret_key_count, TRACE);
     let ctx = assert_ptr_mut!(ctx);
+    let count = assert_ptr_mut!(count);
 
     // We load the keyrings in the background.  But, once TB tries to
     // get the number of certs, we have to wait on that process.
@@ -841,5 +839,5 @@ fn rnp_get_secret_key_count(ctx: *mut RnpContext,
         }
     }).count();
     t!("-> {}", *count);
-    RNP_SUCCESS
+    rnp_success!()
 }
