@@ -1,8 +1,10 @@
 use std::collections::HashSet;
-use std::sync::{Arc, RwLock};
+use std::sync::{Arc, Condvar, Mutex, RwLock};
 use std::time::Duration;
 use std::time::SystemTime;
 
+use anyhow::Result;
+
 use rusqlite::{
     Connection,
     types::ToSql,
@@ -13,7 +15,7 @@ use rusqlite::{
 use sequoia_openpgp as openpgp;
 use openpgp::{Cert, Fingerprint};
 use openpgp::packet::UserID;
-use sequoia_openpgp::policy::Policy;
+use sequoia_openpgp::policy::StandardPolicy;
 use sequoia_wot::{Root, Roots, store::Store};
 
 
@@ -22,17 +24,96 @@ use crate::gpg::Validity;
 use crate::keystore::KeystoreData;
 use crate::tbprofile::TBProfile;
 
+/// Controls tracing in this module.
+const TRACE: bool = crate::TRACE;
+
+/// Handle for communication with the background thread.
+#[derive(Clone)]
+pub struct WoTWorkerHandle {
+    cond: Arc<(Mutex<bool>, Condvar)>,
+}
+
+impl WoTWorkerHandle {
+    /// Constructs a new handle.
+    fn new() -> Self {
+        Self {
+            cond: Arc::new((Mutex::new(false), Condvar::new())),
+        }
+    }
+
+    /// Blocks the current thread until it is notified, or the given
+    /// timeout elapsed.
+    fn wait_for(&self, timeout: Duration) {
+        let (mut notified, _) = self.cond.1.wait_timeout_while(
+            self.cond.0.lock().unwrap(), timeout,
+            |notified| ! *notified).unwrap();
+        *notified = false;
+    }
+
+    /// Wakes the background thread to do a WoT update.
+    ///
+    /// If the background thread is currently doing an update, it
+    /// will, after completing the current update, immediately
+    /// consider doing an update again, unless the result could not
+    /// possibly have changed.
+    pub fn notify(&self) {
+        let mut notified = self.cond.0.lock().unwrap();
+        *notified = true;
+        self.cond.1.notify_one();
+    }
+}
+
 pub struct WoT {
+    /// Handle for communication with the background thread.
+    handle: WoTWorkerHandle,
+
+    ksd: Arc<RwLock<KeystoreData>>,
+    policy: Arc<RwLock<StandardPolicy<'static>>>,
+
     conn: rusqlite::Connection,
     _gpg_ctx: gpg::Ctx,
-    last_update: SystemTime,
+
+    /// Known last_update stamp from the KeyStore.
+    last_keystore_update: usize,
+
+    /// Time at which our WoT calculation possibly go stale and need
+    /// to be recomputed, provided the KeyStore wasn't modified.
+    next_wot_change_at: Option<SystemTime>,
 }
 
 impl WoT {
+    /// Starts the Web-of-Trust background thread.
+    ///
+    /// If we don't have or don't find a Profile directory, or if the
+    /// acceptance database doesn't match our expectations, we will
+    /// emit a warning, ignore the error, and still return a
+    /// `WoTWorkerHandle` which can be notified, but that will have no
+    /// effect.
+    pub fn background_thread_start(ksd: Arc<RwLock<KeystoreData>>,
+                                   policy: Arc<RwLock<StandardPolicy<'static>>>)
+                                   -> Result<WoTWorkerHandle>
+    {
+        let handle = WoTWorkerHandle::new();
+
+        match Self::new(handle.clone(), ksd, policy) {
+            Ok(mut wot) => {
+                std::thread::Builder::new().name("sq wot".to_string())
+                    .spawn(move || wot.background_thread())?;
+            },
+            Err(e) => global_warn!("{}", e),
+        }
+
+        Ok(handle)
+    }
+
     // Return a locked WoTData.
     //
     // If WoT hasn't been initialized, do it now.
-    pub(crate) fn new() -> openpgp::Result<WoT> {
+    fn new(handle: WoTWorkerHandle,
+           ksd: Arc<RwLock<KeystoreData>>,
+           policy: Arc<RwLock<StandardPolicy<'static>>>)
+           -> Result<WoT>
+    {
         let gpg_ctx = gpg::Ctx::new()?;
 
         let path = TBProfile::path()
@@ -57,7 +138,6 @@ impl WoT {
                 let err = err.context(
                     "openpgp.sqlite: \
                      Unexpected schema (querying acceptance_email)");
-                global_warn!("openpgp.sqlite: {}", err);
                 err
             })?;
         conn.query_row("SELECT fpr, decision FROM acceptance_decision LIMIT 1",
@@ -68,7 +148,6 @@ impl WoT {
                 let err = err.context(
                     "openpgp.sqlite:\
                      Unexpected schema (querying acceptance_decision)");
-                global_warn!("openpgp.sqlite: {}", err);
                 err
             })?;
 
@@ -119,44 +198,112 @@ impl WoT {
         tx.commit()?;
 
         Ok(WoT {
+            handle,
+            ksd,
+            policy,
             _gpg_ctx: gpg_ctx,
             conn,
-            last_update: std::time::UNIX_EPOCH,
+            last_keystore_update: 0,
+            next_wot_change_at: None,
         })
     }
 
+    fn background_thread(&mut self) -> ! {
+        rnp_function!(WoT::background_thread, TRACE);
+        loop {
+            let now = SystemTime::now();
+            let t = self.next_wot_change_at.as_ref()
+                .and_then(|t| t.duration_since(now).ok())
+                .unwrap_or(Duration::new(0, 0));
+            t!("Blocking for {:?}", t);
+            self.handle.wait_for(t);
+
+            if let Err(e) = self.update(now) {
+                eprintln!(
+                    "sequoia-octopus: error in WoT background thread: {}", e);
+            }
+        }
+    }
+
     // Do an update of the WoT validities.
     //
     // Carefully update TB's acceptance DB with GPG's validity.
-    pub fn update(&mut self, ksd: Arc<RwLock<KeystoreData>>, p: &dyn Policy) -> openpgp::Result<()>
-    {
-        rnp_function!(WoT::update, super::TRACE);
+    fn update(&mut self, now: SystemTime) -> Result<()> {
+        rnp_function!(WoT::update, TRACE);
 
-        let now = SystemTime::now();
-        let elapsed = now.duration_since(self.last_update);
-        match elapsed {
-            Err(err) => {
-                t!("Error computing time since last update: {}", err);
-                return Ok(());
-            }
-            Ok(elapsed) => {
-                if elapsed < Duration::new(10, 0) {
-                    t!("Throttling updates.");
-                    return Ok(());
-                }
-            }
+        // Read lock the key store, do a scan across the certs, and
+        // release it as soon as possible.
+        let keystore_data = self.ksd.read().unwrap();
+
+        // Check the stored key store timestamp.
+        let last_keystore_update = keystore_data.last_update();
+        if self.last_keystore_update == last_keystore_update
+            && self.next_wot_change_at.map(|t| now < t).unwrap_or(false)
+        {
+            t!("Keystore has not changed, and no known cert component \
+                or signature could have possibly changed, skipping.");
+            return Ok(());
         }
 
-        t!("Checking for WoT updates...");
+        // We compute the next time at which a cert component or
+        // signature is valid or invalid.
+        let mut next_wot_change_at: SystemTime =
+            openpgp::types::Timestamp::from(u32::MAX).into();
 
         // We clone the certificates in the key store to be able to drop the
         // keystore lock as soon as possible.
-        let keystore_data = ksd.read().unwrap();
         let certs: Vec<Cert> = keystore_data.iter()
-            .map(|c| (*c).clone())
+            .map(|c| {
+                // Considers time `t`.
+                let mut consider = |t| {
+                    if t > now && t < next_wot_change_at {
+                        next_wot_change_at = t;
+                    }
+                };
+
+                // All possible certification (sub)keys.
+                for skb in c.keys().filter(|k| k.pk_algo().for_signing()) {
+                    for sig in skb.signatures() {
+                        consider(sig.signature_creation_time().unwrap());
+                        if let Some(t) = sig.signature_expiration_time() {
+                            consider(t);
+                        }
+                        if let Some(t) = sig.key_expiration_time(skb.key()) {
+                            consider(t);
+                        }
+                    }
+                }
+
+                // All possible certification user IDs.
+                for sig in c.userids().flat_map(|u| u.signatures()) {
+                    consider(sig.signature_creation_time().unwrap());
+                    if let Some(t) = sig.signature_expiration_time() {
+                        consider(t);
+                    }
+                    if let Some(t) = sig.key_expiration_time(&c.primary_key()) {
+                        consider(t);
+                    }
+                }
+
+                // All possible certification user attributes.
+                for sig in c.user_attributes().flat_map(|u| u.signatures()) {
+                    consider(sig.signature_creation_time().unwrap());
+                    if let Some(t) = sig.signature_expiration_time() {
+                        consider(t);
+                    }
+                    if let Some(t) = sig.key_expiration_time(&c.primary_key()) {
+                        consider(t);
+                    }
+                }
+
+                (*c).clone()
+            })
             .collect();
+
+        // Drop the lock.
         drop(keystore_data);
 
+        t!("Checking for WoT updates...");
         t!("Keystore size: {}", certs.len());
 
         let reference_time = SystemTime::now();
@@ -190,10 +337,9 @@ impl WoT {
         let roots = Roots::new(trust_roots.clone());
         t!("Trust roots (ultimate ownertrust): {:?}", roots);
 
-        let network = sequoia_wot::Network::from_cert_refs(
-            &certs, &*p, reference_time).unwrap();
-        let mut query =
-            sequoia_wot::Query::new(&network, roots);
+        let policy = (*self.policy.read().unwrap()).clone();
+        let mut network = sequoia_wot::Network::from_cert_refs(
+            &certs, &policy, reference_time, roots).unwrap();
 
         // Add GnuPG intermediate trust roots to 'trust_roots'
         let mut found_one = true;
@@ -234,7 +380,7 @@ impl WoT {
                         continue;
                     }
 
-                    let authenticated_amount = query.authenticate(
+                    let authenticated_amount = network.authenticate(
                         u.userid(), r.fingerprint(), sequoia_wot::FULLY_TRUSTED)
                         .amount();
 
@@ -247,8 +393,9 @@ impl WoT {
                             r.amount());
                         found_one = true;
                         trust_roots.push(r);
-                        query = sequoia_wot::Query::new(
-                            &network, Roots::new(trust_roots.clone()));
+                        let roots = Roots::new(trust_roots.clone());
+                        network = sequoia_wot::Network::from_cert_refs(
+                            &certs, &policy, reference_time, roots).unwrap();
                         continue 'root;
                     } else {
                         t!("Non-ultimately trusted binding <{}, {}> \
@@ -265,14 +412,15 @@ impl WoT {
             }
         }
 
+        let start = SystemTime::now();
         let validity: Vec<(Fingerprint, Vec<(String, Validity)>)> =
             certs.iter().map(|c| {
                 let fp = c.fingerprint();
                 let userids: Vec<(String, Validity)> =
-                    c.with_policy(&*p, reference_time).map(|vc| {
+                    c.with_policy(&policy, reference_time).map(|vc| {
                         vc.userids().filter_map(|uid| {
                             if let Ok(s) = String::from_utf8(uid.value().to_vec()) {
-                                let paths = query.authenticate(
+                                let paths = network.authenticate(
                                     uid.userid(), &fp,
                                     sequoia_wot::FULLY_TRUSTED);
 
@@ -292,10 +440,37 @@ impl WoT {
                     }).unwrap_or(vec![]);
                 (fp, userids)
             }).collect();
+        if let Ok(elapsed) = start.elapsed() {
+            t!("processed {} certs in {:?}, {:?} per cert",
+               validity.len(), elapsed,
+               elapsed.checked_div(validity.len() as u32).unwrap_or(Duration::new(0, 0)));
+        }
 
-        t!("sequoia-wot validities: {:#?}", validity);
-
-        let mut tx = self.conn.transaction()?;
+        // The global authentication was hard to compute, make some
+        // effort to actually write it to the database.  Eagerly start
+        // an immediate transaction (i.e. acquire the write lock right
+        // now) so that we can do all the updates, or fail here, where
+        // we can retry with exponential backoff.
+        let mut tx = {
+            let mut tries = 0;
+            loop {
+                match self.conn.transaction_with_behavior(
+                    rusqlite::TransactionBehavior::Immediate)
+                {
+                    Ok(t) => break t,
+                    Err(e) => {
+                        // Sleep for 3ms, 30ms, 300ms, then fail.
+                        if tries > 2 {
+                            return Err(e.into());
+                        } else {
+                            std::thread::sleep(
+                                Duration::new(0, 3 * 1000 * 10u32.pow(tries)));
+                            tries += 1;
+                        }
+                    },
+                }
+            }
+        };
 
         // See comment in `WoT::new` regarding what exactly is
         // considered managed by sequoia.
@@ -321,8 +496,11 @@ impl WoT {
 
         tx.commit()?;
 
-        self.last_update = SystemTime::now();
+        self.last_keystore_update = last_keystore_update;
+        self.next_wot_change_at = Some(next_wot_change_at);
 
+        t!("last_keystore_update: {}, next_wot_change_at: {:?}",
+           self.last_keystore_update, self.next_wot_change_at);
         Ok(())
     }
 
@@ -334,7 +512,7 @@ impl WoT {
                        validity: Vec<(Fingerprint, Vec<(String, gpg::Validity)>)>)
                        -> openpgp::Result<()>
     {
-        rnp_function!(WoT::import_validity, super::TRACE);
+        rnp_function!(WoT::import_validity, TRACE);
 
         // First, remove all assertions managed by sequoia
         // (they will get re-added below, if they are still valid).
